-- A3Pv2.8/proasic3v3/gategen.vhdl
-- created jeu. août  8 10:18:01 CEST 2019  by Yann Guidon (whygee@f-cpu.org)
-- version sam. août 10 08:06:58 CEST 2019
-- version dim. août 11 17:36:51 CEST 2019
-- version lun. août 12 18:36:04 CEST 2019
-- version mar. août 13 18:05:15 CEST 2019
-- version dim. août 25 10:25:30 CEST 2019 : DLxxxx added
-- version jeu. nov. 21 03:38:34 CET  2019 : removing "trace"vs"fast"
-- version ven. nov. 22 23:39:03 CET  2019 : generic gates
-- version ven. déc. 13 17:31:50 CET  2019 : LUT16
-- version lun. mars 23 04:52:26 CET  2020 : v2.8 adds -ggate_type="simple"
-- version mar. juil. 14 23:15:00 CEST 2020 : v2.9 adds some 4-inputs gates
-- version sam. juil. 18 06:11:26 CEST 2020 : finally spotted a ridiculous
--     and shamefully old mistake in the DFN code. TEST TEST TEST !!!
--
-- Released under the GNU AGPLv3 license
--
-- This file generates PA3_components.vhdl and PA3_definitions.vhdl.
--
-- TODO : reduce countless redundancies...

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.PA3_genlib.all;
use STD.textio.all;
Library SLV;
    use	SLV.SLV_utils.all;

entity GateGen is
  generic ( gate_delay : string := "";
            gate_type  : string := "trace"); -- default architecture name.
end GateGen;

architecture runme of GateGen is
  file gate_defs : text open write_mode is "PA3_definitions.vhdl";
  file gate_list : text open write_mode is "PA3_components.vhdl";
begin

  bench: process
    variable c, Epol, Cpol, Ppol, op : character;
    variable i, j, k, indent : integer := 0;
    variable spaces : string(1 to 12) := (others=>' '); 
    variable s : EnumGates2;
    variable t : EnumGates3;
    variable L2 : LookupType2;
    variable L4 : LookupType4;
    variable L8 : LookupType8;
    variable L16: LookupType16;
    variable row : line;
    variable QN, StrE : string(1 to 2);
    variable GCLK : string(1 to 3);
    variable StrC, StrP : string(1 to 5);
    constant letters: string(1 to 4) := "ABCD";

    procedure wl(s: string) is
    begin
      write(row, s);
      writeline(gate_list, row);
    end wl;

    procedure wd(s: string) is
    begin
      write(row, s);
      writeline(gate_defs, row);
    end wd;

    procedure wdi(s: string) is
    begin
      write(row, spaces(1 to indent) & s);
      writeline(gate_defs, row);
    end wdi;

    procedure wdg is
    begin
      write(row, string'("  generic( "));
      if gate_type /= "simple" then
        wd("Gate_Number : integer := Gate_Census;");
        write(row, string'("           "));
      end if;
      wd("exclude: std_logic_vector := """");");
    end wdg;

    procedure uselib is
    begin
      wd("");
      wd("--------------------------------------------------");
      wd("");
      wd("Library ieee;");
      wd("    use ieee.std_logic_1164.all;");
      if gate_type /= "simple" then
        wd("Library work;");
        wd("    use work.all;");
        wd("    use work.PA3_genlib.all;");
      end if;
      wd("");
    end uselib;

    procedure shift_indent(i : integer) is
    begin
      if i >= 4 then
        indent := i;
      end if;
    end shift_indent;

    procedure recurse_if(level: integer; spaces : string; index: integer) is
    begin
      if level < 1 then
        wd(spaces & c & " <= LUT(" & integer'image(k) & ");");
        k := k+1;
      else
        wd(spaces & "if " & letters(index) & "/='1' then");
          recurse_if(level-1, spaces & "  ", index+1);
        wd(spaces & "else");
          recurse_if(level-1, spaces & "  ", index+1);
        wd(spaces & "end if;");
      end if;
    end recurse_if;

    procedure BCD(i : integer) is
    begin
      if i > 1 then write(row,string'(", B")); end if;
      if i > 2 then write(row,string'(", C")); end if;
      if i > 3 then write(row,string'(", D")); end if;
    end BCD;

  begin
    wl("-- File : PA3_components.vhdl v2.9");
    wl("-- DO NOT MODIFY ! Generated file !");
    wl("-- (C) Yann Guidon 20190808 - 20200718");
    wl("");
    wl("Library ieee;");
    wl("    use ieee.std_logic_1164.all;");
    wl("");
    wl("package proasic3 is");

    wd("-- File : PA3_definitions.vhdl v2.9");
    wd("-- DO NOT MODIFY ! Generated file !");
    wd("-- (C) Yann Guidon 20190808 - 20200718");
    if gate_delay = "" then
      wd("--  * This version has no delay.");
      report "No gate delay specified. Change this value with -ggate_delay=""1 ns"" ";
    else
      wd("--  * This version has delay set to " & gate_delay);
      report "Gate delay set to " & gate_delay;
    end if;

    if gate_type = "simple" then
      wd("--  * Gate type set to 'simple', advanced analysis is disabled.");
      report "Gate type set to 'simple', advanced analysis is disabled.";
      if gate_delay /= "" then
        wd("--  * Warning : MX2x behaviour might have incorrect timing");
        report "Warning : Gate delay not null => MX2x behaviour might be incorrect";
      end if;
    else
      wl("");
      wl("  component register_generics generic (gate_select_number, bit_select_number,");
      wl("      flip_LUT : integer; verbose, filename : string);");
      wl("      port(clock : in std_logic); end component;"); 
      wl("");

      uselib;
      wd("-- Note if you use the modes trace, alter, netlist :");
      wd("-- /!\ You MUST write this in front/before including or instanciating the other gates !");
      wd("-- It must be called first to initialise important values from the generics /!\");
      wd("entity register_generics is");
      wd("  generic( gate_select_number : integer :=  0;");
      wd("            bit_select_number : integer := -1;");
      wd("                      flip_LUT: integer := 0;");
      wd("            verbose, filename : string := """");");
      wd("  port(clock : in std_logic);");
      wd("end register_generics;");
      wd("");
      wd("architecture rtl of register_generics is");
      wd("begin");
      wd("  update_select_gate(gate_select_number, bit_select_number,");
      wd("   flip_LUT, verbose, filename); -- called only once at start");
      wd("");
      wd("  process(clock) begin");
      wd("    case clock is");
      wd("      when '1' => update_histogram;");
      wd("      when 'X' => display_histogram;");
      wd("      when others => NULL;");
      wd("    end case;");
      wd("  end process;");
      wd("end rtl;");

      wd("");
      wd("--------------------------------------------------");
    end if;


    wd("");
    wd("-- Generic gates :");

    j := 2;
    for i in 1 to 4 loop
      write(row,"  component generic_gate" & integer'image(i) & " generic(exclude: std_logic_vector); port(A");
      if i > 1 then  write(row,string'(", B")); else  write(row,string'("   ")); end if;
      if i > 2 then  write(row,string'(", C")); else  write(row,string'("   ")); end if;
      if i > 3 then  write(row,string'(", D")); else  write(row,string'("   ")); end if;
      wl(" : in std_logic; LUT : in std_logic_vector; Y : out std_logic); end component;");

      uselib;
      wd("entity generic_gate" & integer'image(i) & " is");
      wdg;
      write(row,string'("  port( A"));
      BCD(i);
      wd(" : in std_logic;");
      wd("        LUT : in std_logic_vector;");
      wd("        Y : out std_logic);");
      wd("end generic_gate" & integer'image(i) & ";");
      wd("");
      wd("architecture " & gate_type & " of generic_gate" & integer'image(i) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT" & integer'image(j) & " : LookupType" & integer'image(j) & ";");
        wd("begin");
        wd("  LUT" & integer'image(j) & " <= InferLUT(LUT, LUT" & integer'image(j)
             &  "'instance_name, Gate_Number, "  & integer'image(j-1) & ", exclude);");    -- called only once at start
        write(row, string'("  Y <= LookItUp") & integer'image(i) & string'("(A"));
        BCD(i);
        write(row, string'(", LUT") & integer'image(j) & string'(", Gate_Number)"));
        if gate_delay /= "" then
          write(row, " after " & gate_delay);
        end if;
        wd(";");
      else
        c := 'Y';
        if gate_delay /= "" then
          wd("  signal S : std_logic;");
          c := 'S';
        end if;

        wd("begin");
        write(row, string'("  process(A"));
        BCD(i);
        wd(", LUT) begin");
        k := 0;
        recurse_if(i, "    ", 1);
        wd("  end process;");

        if gate_delay /= "" then
          wd("  Y <= S after " & gate_delay & ";");
        end if;
      end if;

      wd("end " & gate_type & ";");

      j := j*2;
    end loop;


    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 0-input gates :");

    wl("");
    wl("-- 0-input gates :");

    for i in NoInputGates'range loop
      wl("  component "& NoInputGates(i) & " generic(exclude: std_logic_vector); port(Y : out std_logic); end component;");

      uselib;
      wd("entity " & NoInputGates(i) & " is");
      wdg;
      wd("  port(Y : out std_logic);");
      wd("end " & NoInputGates(i) & ";");

      wd("");
      wd("architecture " & gate_type & " of " & NoInputGates(i) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT1 : LookupType1;");
        wd("begin");
        wd("  LUT1 <= InferLUT(""" & integer'image(i) & integer'image(i)
                 &  """, LUT1'instance_name, Gate_Number, 0, exclude);");
        wd("  Y <= LUT1(0);");
      else
        wd("begin");
        wd("  Y <= '" & integer'image(i) & "';");
      end if;
      wd("end " & gate_type & ";");
    end loop;

    wl("");
    wl("-- 1-input gates :");

    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 1-input gates :");
    
    for i in BufferGates'range loop
      wl("  component "& BufferGates(i) & " generic(exclude: std_logic_vector); port(A : in std_logic; Y : out std_logic); end component;");

      if BufferGates(i)(1) = 'I' then
        L2:="10HL"; -- inverter gate
      else
        L2:="01LH"; -- note: ascending order of the bits !
      end if;
      uselib;
      wd("entity " & trim_spaces(BufferGates(i)) & " is");
      wdg;
      wd("  port( A : in std_logic;");
      wd("        Y : out std_logic);");
      wd("end " & trim_spaces(BufferGates(i)) & ";");

      wd("");
      wd("architecture " & gate_type & " of " & trim_spaces(BufferGates(i)) & " is");

      if gate_type /= "simple" then
        wd("  signal LUT2 : LookupType2;");
        wd("begin");
        wd("  LUT2 <= InferLUT(""" & SLV_to_bin(L2)
           & """, LUT2'instance_name, Gate_Number, 1, exclude);");
        write(row, string'("  Y <= LookItUp1(A, LUT2, Gate_Number)"));
      else
        wd("begin");
        write(row, string'("  Y <= "));
        if BufferGates(i)(1) = 'I' then
          write(row, string'("not "));
        end if;
        write(row, string'("A"));
      end if;
      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");

      wd("end " & gate_type & ";");
    end loop;


    wl("");
    wl("-- 2-inputs gates :");

    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 2-inputs gates :");

    for s in EnumGates2 loop
      L4 := ComputeLUT4(s);
      write(row, "  component " & EnumGates2'image(s), left, 19);
      wl(" generic(exclude: std_logic_vector); port(A, B : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L4(0 to 3)));

      uselib;
      wd("entity " & EnumGates2'image(s) & " is");
      wdg;
      wd("  port( A, B : in std_logic;");
      wd("           Y : out std_logic);");
      wd("end " & EnumGates2'image(s) & ";");

      wd("");
      wd("architecture " & gate_type & " of " & EnumGates2'image(s) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT4 : LookupType4;");
        wd("begin");
        wd("  LUT4 <= InferLUT(""" & SLV_to_bin(L4)
           & """, LUT4'instance_name, Gate_Number, 3, exclude);");
        write(row, string'("  Y <= LookItUp2(A, B, LUT4, Gate_Number)"));
      else
        wd("begin");
	write(row, string'("  Y <= ") & trim_spaces(DefGate2Array(s)) );
      end if;

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");

      wd("end " & gate_type & ";");
    end loop;


    wl("");
    wl("-- 3-inputs gates :");
    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 3-inputs gates :");

    for t in EnumGates3 loop
      case t is
        when MX2  |
             MX2A |
             MX2B |
             MX2C   =>  op := 'S';
        when others =>  op := 'C';
      end case;

      L8 := ComputeLUT8(t);
      write(row, "  component " & EnumGates3'image(t), left, 19);
      wl(" generic(exclude: std_logic_vector); port(A, B, " & op
          & " : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L8(0 to 7)));

      uselib;
      wd("entity " & EnumGates3'image(t) & " is");
      wdg;
      wd("  port( A, B, " & op &" : in std_logic;");
      wd("              Y : out std_logic);");
      wd("end " & EnumGates3'image(t) & ";");

      wd("");
      wd("architecture " & gate_type & " of " & EnumGates3'image(t) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT8 : LookupType8;");
        wd("begin");
        wd("  LUT8 <= InferLUT(""" & SLV_to_bin(L8)
            & """, LUT8'instance_name, Gate_Number, 7, exclude);");    -- called only once at start
        write(row, string'("  Y <= LookItUp3(A, B, ") & op & string'(", LUT8, Gate_Number)"));
      else
	wd("begin");
        write(row, string'("  Y <= ") & trim_spaces(DefGate3Array(t)) );
      end if;
      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("end " & gate_type & ";");
    end loop;


    wl("");
    wl("-- 4-inputs gates :");
    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- 4-inputs gates :");

    for t in EnumGates4 loop

      L16 := ComputeLUT16(t);
      write(row, "  component " & EnumGates4'image(t), left, 20);
      wl(" generic(exclude: std_logic_vector); port(A, B, C, D :"
         & " in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L16(0 to 15)));

      uselib;
      wd("entity " & EnumGates4'image(t) & " is");
      wdg;
      wd("  port( A, B, C, D : in std_logic;");
      wd("                 Y : out std_logic);");
      wd("end " & EnumGates4'image(t) & ";");

      wd("");
      wd("architecture " & gate_type & " of " & EnumGates4'image(t) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT16 : LookupType16;");
        wd("begin");
        wd("  LUT16 <= InferLUT(""" & SLV_to_bin(L16)
            & """, LUT16'instance_name, Gate_Number, 15, exclude);");    -- called only once at start
        write(row, string'("  Y <= LookItUp4(A, B, C, D, LUT16, Gate_Number)"));
      else
	wd("begin");
        write(row, string'("  Y <= ") & trim_spaces(DefGate4Array(t)) );
      end if;
      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("end " & gate_type & ";");
    end loop;


    wl("");
    wl("-- Sequential gates :");
    wd("");
    wd("--------------------------------------------------");
    wd("");
    wd("-- Sequential gates :");

    for i in EnumDFF'range loop

      if EnumDFF(i)(2)='F' then
        GCLK := "CLK";
      else
        GCLK := "G  ";
      end if;

      if EnumDFF(i)(3) = 'I' then
        QN := "QN";
        L2:="10LH"; -- inverter
      else
        QN := " Q";
        L2:="01HL";
      end if;

      StrE := "  ";
      j := 5;
      if EnumDFF(i)(5) = 'E' then
        StrE := "E,";
        j := 7;
        Epol := EnumDFF(i)(6);
      end if;

      StrP := "     ";
      if EnumDFF(i)(j) = 'P' then
        StrP := ", PRE";
        Ppol := EnumDFF(i)(j+1);
        j := j+2;
      end if;

      StrC := "     ";
      if j < 8 and EnumDFF(i)(j) = 'C' then
        StrC := ", CLR";
        Cpol := EnumDFF(i)(j+1);
      end if;

      wl("  component " & EnumDFF(i) & " generic(exclude: std_logic_vector); port(D, " & StrE & " " & GCLK & StrP & StrC
         & " : in std_logic; " & QN &" : out std_logic); end component;");

      uselib;
      wd("entity " & trim_spaces(EnumDFF(i)) & " is");
      wdg;
      wd("  port(D, " & StrE & " " & GCLK & StrP & StrC & " : in std_logic;");
      wd("       " & QN & " : out std_logic);");
      wd("end " & trim_spaces(EnumDFF(i)) & ";");
      wd("");
      wd("architecture " & gate_type & " of " & trim_spaces(EnumDFF(i)) & " is");
      if gate_type /= "simple" then
        wd("  signal LUT2 : LookupType2;");
      end if;
      wd("begin");
      if gate_type /= "simple" then
      wd("  LUT2 <= InferLUT(""" & SLV_to_bin(L2)
         & """, LUT2'instance_name, Gate_Number, 1, exclude);");
      end if;

      write(row, string'("  process(" & GCLK));
      if StrE(1) = 'E' then
        write(row, string'(", E"));
      end if;
      if StrP(1) = ',' then
        write(row, StrP);
      end if;
      if StrC(1) = ',' then
        write(row, StrC);
      end if;
      wd(")");
      wd("  begin");

      shift_indent(4);

      if StrC(1) = ',' then
        wdi("if CLR='" & Cpol & "' then");
        write(row, spaces(1 to indent) & "  "& QN &" <= '0'");
        if gate_delay /= "" then
          write(row, " after " & gate_delay);
        end if;
        wd(";");
        wdi("else");
        shift_indent(6);
      end if;

        -- Clear has precedence over Preset
        if StrP(1) = ',' then
          wdi("if PRE='" & Ppol & "' then");
          write(row, spaces(1 to indent) & "  "& QN &" <= '1'");
          if gate_delay /= "" then
            write(row, " after " & gate_delay);
          end if;
          wd(";");
          wdi("else");
          shift_indent(indent+2);
        end if;

          write(row, spaces(1 to indent) & "if ");

          if EnumDFF(i)(2)='F' then -- Flip Flop
            if EnumDFF(i)(4)='1' then
              write(row, string'("rising"));
            else
              write(row, string'("falling"));
            end if;
            write(row, string'("_edge(CLK)"));
            if StrE(1) = 'E' then
              write(row, " and E='" & Epol & "'");
            end if;
          else         -- Latch
            write(row, "G='" & EnumDFF(i)(4) & "'");
          end if;

          wd(" then");
          write(row, spaces(1 to indent) & "  " & QN & string'(" <= ") );

          if gate_type /= "simple" then
            write(row, string'("LookItUp1(D , LUT2, Gate_Number)"));
          else
            write(row, string'("D"));
          end if;

          if gate_delay /= "" then
            write(row, " after " & gate_delay);
          end if;
          wd(";");
          wdi("end if;");

        if StrP(1) = ',' then
          shift_indent(indent-2);
          wdi("end if;");
        end if;

      if StrC(1) = ',' then
        shift_indent(indent-2);
        wdi("end if;");
      end if;

      wd("  end process;");
      wd("end " & gate_type & ";");
    end loop;


    wl("");
    wl("");
    wl("  component RAM4K9 port(");
    wl("    ADDRA11, ADDRA10, ADDRA9, ADDRA8, ADDRA7, ADDRA6,");
    wl("    ADDRA5,  ADDRA4,  ADDRA3, ADDRA2, ADDRA1, ADDRA0,");
    wl("    ADDRB11, ADDRB10, ADDRB9, ADDRB8, ADDRB7, ADDRB6,");
    wl("    ADDRB5,  ADDRB4,  ADDRB3, ADDRB2, ADDRB1, ADDRB0,");
    wl("    DINA8, DINA7, DINA6, DINA5, DINA4, DINA3, DINA2, DINA1, DINA0,");
    wl("    DINB8, DINB7, DINB6, DINB5, DINB4, DINB3, DINB2, DINB1, DINB0,");
    wl("    WIDTHA0, WIDTHA1, WIDTHB0, WIDTHB1, PIPEA, PIPEB, WMODEA, WMODEB,");
    wl("    BLKA, BLKB, WENA, WENB, CLKA, CLKB, RESET : in std_logic := 'U';");
    wl("    DOUTA8, DOUTA7, DOUTA6, DOUTA5, DOUTA4, DOUTA3, DOUTA2, DOUTA1, DOUTA0,");
    wl("    DOUTB8, DOUTB7, DOUTB6, DOUTB5, DOUTB4, DOUTB3, DOUTB2, DOUTB1, DOUTB0 : out std_logic);");
    wl("  end component;");
    wl("");
    wl("end proasic3;");

    wait;
  end process;

end runme;
