#!/bin/bash
# file A3Pv2.3/test5_ALU8/test.sh
# created by Yann Guidon / whygee@f-cpu.org
# version ven. nov. 10 03:25:21 CET 2017
# version lun. nov. 13 02:36:33 CET 2017
# version jeu. nov. 16 03:07:00 CET 2017
# version lun. avril 1 02:58:19 CEST 2019
# version sam. nov. 23 20:22:07 CET 2019 : integration in A3P library

echo
echo  'This script runs exhaustive self-tests and will take "a while".'
echo  'You may interrupt it (CTRL+C) if needed. Please report errors.'
echo

if [ ! -f ../../proasic3v3/proasic3-obj93.cf  ]; then
  echo -n '--> going to ' ; pushd ../../proasic3v3/
  ./build.sh
  echo -n '<-- back to ' ; popd
fi

if [ ! -f ../Y8/ygrec8-obj93.cf  ]; then
  echo -n '--> going to ' ; pushd ../Y8/
  ./build.sh
  echo -n '<-- back to ' ; popd
fi

#VER="--std=93"
PA3="-P../../proasic3v3 -P../Y8"

run_test () {
# $1 = MSG_name
# $2 = ALU file
# $3 = DEC file
  rm -f alu8 alu8_tb work-obj93.cf *.o &&
  echo -n "ALU $1 standalone: " &&
  ghdl -a $PA3 gate_CLA3.vhdl $3 $2 &&
  ghdl -e $PA3 ALU8 &&
  ./alu8 &&
  echo OK &&
  echo -n "ALU $1 testbench:  " &&
  ghdl -a $PA3 ALU8_tb.vhdl &&
  ghdl -e $PA3 ALU8_tb &&
  ./alu8_tb &&
  echo OK
}

# Available decoders :
#../DEC/dec_ALU8_RTL_bool.vhdl  => simple version, ok pour simulation
##../DEC/dec_ALU8_fast_bool.vhdl => minimal logic, basic equations
#../DEC/dec_ALU8_fast_A3P.vhdl  => minimal and fastest logic, mapped to A3P but toggles a lot
##../DEC/dec_ALU8_latch_RTL.vhdl => version with toggle reduction
##../DEC/dec_ALU8_latch_bool.vhdl =>
### Some versions are not included in the A3P test suite.
### Check in the YGREC8 source archives if you need them.

# Available ALU versions:
# ALU8_classic.vhdl => basic version using standard adder
# ALU8_CLA.vhdl     => radix-3 Carry Lookahead Adder
# ALU8_A3P.vhdl     => like above and mapped to A3P tiles

run_test "simple bool"    "ALU8_classic.vhdl" "dec_ALU8_RTL_bool.vhdl" &&
run_test "CLA bool"       "ALU8_CLA.vhdl"     "dec_ALU8_RTL_bool.vhdl" &&
run_test "A3P bool"       "ALU8_A3P.vhdl"     "dec_ALU8_RTL_bool.vhdl" &&
run_test "A3P fast_A3P"   "ALU8_A3P.vhdl"     "dec_ALU8_fast_A3P.vhdl" &&

rm -f alu8 alu8_tb work-obj93.cf *.o gatelist.log &&

echo &&
echo "OK now let's probe the gates and see if we catch all the errors :" &&
ghdl -a $PA3 gate_CLA3.vhdl dec_ALU8_fast_A3P.vhdl ALU8_A3P.vhdl ALU8_tb.vhdl &&
ghdl -e $PA3 ALU8_tb &&
./alu8_tb -gflip=0 -ggate_select_number=-1 > gatelist.log &&
DEPTH_LIST=$(grep '  gate#' gatelist.log |sed 's/.*(trace):lut//' ) &&
echo "Size of LUTs: " $DEPTH_LIST &&
k=1 && # first gate is n°1
l=0 &&
echo &&
rm -f failures.log &&
# scan the list to extract the depth
for i in $DEPTH_LIST ; do
  # don't process the decoder
  if (( $k > 8 )); then
    # for each bit in the LUT:
    for j in $(seq 0 $((i-1)) ) ; do
      echo -n $'\r'"$l/535  " >&2
      echo 'Run #'$l  'Gate #'$k 'LUT'$i'['$j']'
      (./alu8_tb -gflip=1 -ggate_select_number=$k -gbit_select_number=$j &&
          echo 'failure not found :-(' ) || echo 'Failure detected : great !'
      echo ""
      l=$((l+1))
    done;
  fi
  k=$((k+1))
done >  failures.log &&
echo 'done !' &&

( grep 'failure not found :-(' failures.log &&
    echo "Oh, some failures have not been detected") ||
      ( echo "" &&
        s=0 &&
        for i in $( grep 's:(report note): Error at cycle#' failures.log  | sed 's/.* Error at cycle#//' |sed 's/ .*//' )
        do
          s=$(( s + i ))
        done &&
        echo "All the bits have been successfully checked in $s cycles !" &&
         rm -f alu8 alu8_tb work-obj93.cf *.o ) &&

echo
echo "Finally let's extract the netlist :" &&

ghdl -a $PA3 gate_CLA3.vhdl dec_ALU8_fast_A3P.vhdl ALU8_A3P.vhdl Wrap_ALU8.vhdl ../../proasic3v3/VectGen.vhdl &&
ghdl -e $PA3 VectGen &&
./vectgen  | sed 's/.*te)://'

